"""
Test the apps' views
"""

# Standard Library
import json
from http import HTTPStatus
from unittest.mock import patch

# Third Party
from requests import RequestException

# Django
from django.test import TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_fake_user

# AA ESI Status
from esistatus import views


class TestDashboardWidget(TestCase):
    """
    Test the dashboard widget
    """

    def setUp(self):
        """
        Set up users

        :return:
        :rtype:
        """

        self.superuser = create_fake_user(
            character_id=1002, character_name="Clark Kent"
        )
        self.superuser.is_superuser = True
        self.superuser.is_staff = True

        self.normal_user = create_fake_user(
            character_id=1001, character_name="Peter Parker"
        )

    # def test_dashboard_widget_with_superuser(self):
    #     self.client.force_login(self.superuser)
    #     response = self.client.get(path=reverse(viewname="authentication:dashboard"))
    #     self.assertContains(
    #         response,
    #         '<div class="card-title text-center mb-0">ESI Status</div>',
    #         html=True,
    #     )

    def test_dashboard_widget_with_normal_user(self):
        """
        Test that a normal user does not see the ESI status widget

        :return:
        :rtype:
        """

        self.client.force_login(user=self.normal_user)
        response = self.client.get(path=reverse(viewname="authentication:dashboard"))
        self.assertNotContains(
            response=response,
            text='<div class="card-title text-center mb-0">ESI Status</div>',
            html=True,
        )


class TestAjaxEsiStatus(TestCase):
    """
    Test the AJAX ESI Status view
    """

    def setUp(self):
        """
        Set up users

        :return:
        :rtype:
        """

        self.superuser = create_fake_user(
            character_id=1002, character_name="Clark Kent"
        )
        self.superuser.is_superuser = True

        self.client.force_login(user=self.superuser)

    @patch("esistatus.views._esi_status")
    def test_ajax_esi_status_with_status_result(self, mock_esi_status):
        """
        Test the AJAX ESI Status view with HTTPStatus.OK

        :param mock_esi_status:
        :type mock_esi_status:
        :return:
        :rtype:
        """

        mock_esi_status.return_value = (True, {})
        response = self.client.get(path=reverse(viewname="esistatus:ajax_esi_status"))
        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)

    @patch("esistatus.views._esi_status")
    def test_ajax_esi_status_without_status_result(self, mock_esi_status):
        """
        Test the AJAX ESI Status view with HTTPStatus.NO_CONTENT

        :param mock_esi_status:
        :type mock_esi_status:
        :return:
        :rtype:
        """

        mock_esi_status.return_value = (False, {})
        response = self.client.get(path=reverse(viewname="esistatus:ajax_esi_status"))
        self.assertEqual(first=response.status_code, second=HTTPStatus.NO_CONTENT)


class TestIndex(TestCase):
    """
    Test the index view
    """

    def setUp(self):
        """
        Set up users

        :return:
        :rtype:
        """

        self.superuser = create_fake_user(
            character_id=1002, character_name="Clark Kent"
        )
        self.superuser.is_superuser = True

        self.client.force_login(user=self.superuser)

    @patch("esistatus.views._esi_status")
    def test_index_with_status_result(self, mock_esi_status):
        """
        Test the index view with a status result

        :param mock_esi_status:
        :type mock_esi_status:
        :return:
        :rtype:
        """

        mock_esi_status.return_value = (True, {})
        response = self.client.get(path=reverse(viewname="esistatus:index"))

        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)

    @patch("esistatus.views._esi_status")
    def test_index_without_status_result(self, mock_esi_status):
        """
        Test the index view without a status result

        :param mock_esi_status:
        :type mock_esi_status:
        :return:
        :rtype:
        """

        mock_esi_status.return_value = (False, {})
        response = self.client.get(path=reverse(viewname="esistatus:index"))

        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)


class TestEsiStatus(TestCase):
    """
    Test the _esi_status function
    """

    @patch("requests.get")
    def test_esi_status_with_valid_response(self, mock_get):
        """
        Test the _esi_status function with a valid response

        :param mock_get:
        :type mock_get:
        :return:
        :rtype:
        """

        mock_get.return_value.json.return_value = []
        mock_get.return_value.raise_for_status.return_value = None
        has_status_result, esi_endpoint_status = views._esi_status()

        self.assertTrue(has_status_result)
        self.assertEqual(
            first=esi_endpoint_status,
            second={
                "green": {"endpoints": {}, "count": 0, "percentage": "0.00%"},
                "yellow": {"endpoints": {}, "count": 0, "percentage": "0.00%"},
                "red": {"endpoints": {}, "count": 0, "percentage": "0.00%"},
            },
        )

    @patch("requests.get")
    def test_esi_status_with_request_exception(self, mock_get):
        """
        Test the _esi_status function with a request exception

        :param mock_get:
        :type mock_get:
        :return:
        :rtype:
        """

        mock_get.side_effect = RequestException("Request Exception")
        has_status_result, esi_endpoint_status = views._esi_status()

        self.assertFalse(expr=has_status_result)
        self.assertEqual(first=esi_endpoint_status, second={})

    @patch("requests.get")
    def test_esi_status_with_json_decode_error(self, mock_get):
        """
        Test the _esi_status function with a JSON decode error

        :param mock_get:
        :type mock_get:
        :return:
        :rtype:
        """

        mock_get.return_value.json.side_effect = json.JSONDecodeError(
            msg="Expecting property name enclosed in double quotes", doc="{}", pos=0
        )
        has_status_result, esi_endpoint_status = views._esi_status()

        self.assertFalse(expr=has_status_result)
        self.assertEqual(first=esi_endpoint_status, second={})



"""
App url config
"""

# Django
from django.urls import path

# AA ESI Status
from esistatus import views

app_name: str = "esistatus"

urlpatterns = [
    path(route="", view=views.index, name="index"),
    path(route="ajax/esi_status/", view=views.ajax_esi_status, name="ajax_esi_status"),
]



"""
Test checks for access to fleetpings
"""

# Standard Library
from http import HTTPStatus

# Django
from django.contrib.auth.models import Group
from django.test import TestCase
from django.urls import reverse

# AA ESI Status
from esistatus.tests.utils import create_fake_user


class TestAccess(TestCase):
    """
    Test access to the app
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()

        cls.group = Group.objects.create(name="Superhero")

        cls.user_1001 = create_fake_user(
            character_id=1001, character_name="Peter Parker"
        )
        cls.user_1002 = create_fake_user(character_id=1002, character_name="Clark Kent")
        cls.user_1002.is_superuser = True
        cls.user_1003 = create_fake_user(
            character_id=1003, character_name="Bruce Wayne"
        )
        cls.user_1003.is_staff = True

    def test_has_access(self):
        """
        Test that a user without access gets a 302
        """

        # when
        res = self.client.get(path=reverse(viewname="esistatus:index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_logged_in_has_access(self):
        """
        Test that a user with access gets to see it
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="esistatus:index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_superuser_has_access(self):
        """
        Test that a superuser has access
        """

        # given
        self.client.force_login(user=self.user_1002)

        # when
        res = self.client.get(path=reverse(viewname="esistatus:index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_staff_user_has_access(self):
        """
        Test that a staff user has access
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(path=reverse(viewname="esistatus:index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_response_content(self):
        """
        Test that the response contains the expected content
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="esistatus:index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text="ESI Status")



"""
The models
"""

# Django
from django.db import models
from django.utils.translation import gettext_lazy as _


class EsiStatus(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        managed = False
        default_permissions = ()
        permissions = (("basic_access", _("Can access this app")),)



"""
The views
"""

# Standard Library
import json
from typing import Any, Dict, Tuple

# Third Party
import requests

# Django
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import render
from django.template.loader import render_to_string

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA ESI Status
from esistatus import __title__
from esistatus.constants import USER_AGENT

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def _append_value(dict_obj: Dict, key: str, value: Any) -> None:
    """
    Append a value to a key in a dictionary

    :param dict_obj: The dictionary
    :type dict_obj: Dict
    :param key: The key
    :type key: str
    :param value: The value
    :type value: Any
    :return: None
    :rtype: None
    """

    # Check if key exists in the dict or not
    if key in dict_obj:
        # Key exist in the dict.
        # Check if the type of the value of a key is a list or not
        if not isinstance(dict_obj[key], list):
            # If the type is not list then make it list
            dict_obj[key] = [dict_obj[key]]

        # Append the value in a list
        dict_obj[key].append(value)
    else:
        # If the key is not in the dict, add a key-value pair
        dict_obj[key] = [value]


def _esi_endpoint_status_from_json(esi_endpoint_json: json) -> Tuple:
    """
    Get the ESI endpoint status from the ESI json

    :param esi_endpoint_json: The ESI endpoint json
    :type esi_endpoint_json: json
    :return: The ESI endpoint status
    :rtype: Tuple[(bool) Has Status Results, (dict) Endpoint Status]
    """

    esi_endpoint_status = {
        "green": {"endpoints": {}, "count": 0, "percentage": "0.00%"},
        "yellow": {"endpoints": {}, "count": 0, "percentage": "0.00%"},
        "red": {"endpoints": {}, "count": 0, "percentage": "0.00%"},
    }

    for esi_endpoint in esi_endpoint_json:
        _append_value(
            dict_obj=esi_endpoint_status[esi_endpoint["status"]]["endpoints"],
            key=esi_endpoint["tags"][0],
            value={
                "route": esi_endpoint["route"],
                "method": esi_endpoint["method"].upper(),
            },
        )

        esi_endpoint_status[esi_endpoint["status"]]["count"] += 1

    endpoints_total = (
        esi_endpoint_status["green"]["count"]
        + esi_endpoint_status["yellow"]["count"]
        + esi_endpoint_status["red"]["count"]
    )

    # Calculate percentages - Green endpoints
    green_percentage_calc = (
        (esi_endpoint_status["green"]["count"] / endpoints_total * 100)
        if esi_endpoint_status["green"]["count"] > 0
        else 0
    )
    esi_endpoint_status["green"]["percentage"] = f"{green_percentage_calc:.2f}%"

    # Calculate percentages - Yellow endpoints
    yellow_percentage_calc = (
        (esi_endpoint_status["yellow"]["count"] / endpoints_total * 100)
        if esi_endpoint_status["yellow"]["count"] > 0
        else 0
    )
    esi_endpoint_status["yellow"]["percentage"] = f"{yellow_percentage_calc:.2f}%"

    # Calculate percentages - Red endpoints
    red_percentage_calc = (
        (esi_endpoint_status["red"]["count"] / endpoints_total * 100)
        if esi_endpoint_status["red"]["count"] > 0
        else 0
    )
    esi_endpoint_status["red"]["percentage"] = f"{red_percentage_calc:.2f}%"

    # Return the whole jazz (Tuple[(bool) Has Status Results, (dict) Endpoint Status])
    return True, esi_endpoint_status


def _esi_status() -> Tuple:
    """
    Get the ESI status

    :return: The ESI status
    :rtype: Tuple (Tuple[(bool) Has Status Results, (dict) Endpoint Status])
    """

    has_status_result = False
    request_headers = {"User-Agent": USER_AGENT}
    esi_status_json_url = "https://esi.evetech.net/status.json?version=latest"
    esi_endpoint_status = {}

    try:
        esi_endpoint_status_result = requests.get(
            url=esi_status_json_url, headers=request_headers, timeout=10
        )
        esi_endpoint_status_result.raise_for_status()
        esi_endpoint_json = esi_endpoint_status_result.json()
    except requests.exceptions.RequestException as exc:
        error_str = str(exc)

        logger.info(msg=f"Unable to get ESI status. Error: {error_str}")

        return has_status_result, esi_endpoint_status
    except json.JSONDecodeError:
        logger.info(
            msg=(
                "Unable to get ESI status. ESI returning gibberish, I can't understand …"
            )
        )

        return has_status_result, esi_endpoint_status

    has_status_result, esi_endpoint_status = _esi_endpoint_status_from_json(
        esi_endpoint_json=esi_endpoint_json
    )

    return has_status_result, esi_endpoint_status


def index(request: WSGIRequest) -> HttpResponse:
    """
    Index view

    :param request: The request
    :type request: WSGIRequest
    :return: The response
    :rtype: HttpResponse
    """

    has_status_result, esi_endpoint_status = _esi_status()

    context = {
        "has_status_result": has_status_result,
        "esi_endpoint_status": esi_endpoint_status,
    }

    return render(
        request=request, template_name="esistatus/index.html", context=context
    )


def ajax_esi_status(request: WSGIRequest) -> HttpResponse:
    """
    AJAX ESI Status view

    :param request: The request
    :type request: WSGIRequest
    :return: The response
    :rtype: HttpResponse
    """

    has_status_result, esi_endpoint_status = _esi_status()

    if has_status_result:
        context = {
            "has_status_result": has_status_result,
            "esi_endpoint_status": esi_endpoint_status,
        }

        return render(
            request=request,
            template_name="esistatus/partials/dashboard-widget/esi-status.html",
            context=context,
        )

    return HttpResponse(status=204)


def dashboard_widget(request: WSGIRequest) -> str:
    """
    Dashboard widget

    :param request: The request
    :type request: WSGIRequest
    :return: The widget
    :rtype: str
    """

    if request.user.is_superuser:
        return render_to_string(
            template_name="esistatus/dashboard-widget.html", request=request
        )

    return ""



"""
Hook into AA
"""

# Alliance Auth
from allianceauth import hooks
from allianceauth.hooks import DashboardItemHook
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA ESI Status
from esistatus import __title__, urls
from esistatus.views import dashboard_widget


class AaEsiStatusMenuItem(MenuItemHook):  # pylint: disable=too-few-public-methods
    """
    This class ensures only authorized users will see the menu entry
    """

    def __init__(self):
        # Setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            text=__title__,
            classes="fa-solid fa-signal",
            url_name="esistatus:index",
            navactive=["esistatus:"],
        )

    def render(self, request):
        """
        Check if the user has the permission to view this app

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return MenuItemHook.render(self, request)


class AaEsiStatusDashboardHook(DashboardItemHook):
    """
    This class adds the widget to the dashboard
    """

    def __init__(self):
        # Setup dashboard widget
        DashboardItemHook.__init__(self=self, view_function=dashboard_widget, order=1)


@hooks.register("menu_item_hook")
def register_menu():
    """
    Register our menu item

    :return: The hook
    :rtype: AaEsiStatusMenuItem
    """

    return AaEsiStatusMenuItem()


@hooks.register("url_hook")
def register_urls():
    """
    Register our base url

    :return: The hook
    :rtype: UrlHook
    """

    return UrlHook(
        urls=urls,
        namespace="esistatus",
        base_url=r"^esi-status/",
        excluded_views=["esistatus.views.index"],
    )


@hooks.register("dashboard_hook")
def register_esi_hook():
    """
    Register our dashboard hook

    :return: The hook
    :rtype: AaEsiStatusDashboardHook
    """

    return AaEsiStatusDashboardHook()
